/*
 * © Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <iostream>
#include <string>
#include <sstream>
#include <istream>
#include <algorithm>
#include <iterator>
#include <infiniband/misc/regexp/regexp.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>


#include "ibdiagpath.h"

#include "ibdiagnet_debug_wrap.h"

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h>

/****************************************************/
/* globals */
/****************************************************/


/****************************************************/
/*******************public methods*******************/
/****************************************************/

IBDiagpath::IBDiagpath()
{
}

Stage* IBDiagpath::CreateStage(uint8_t stageIndex) {

    IBDIAGNET_ENTER;
    Stage *stage = NULL;
    switch (stageIndex) {
        case STAGE_IDX_DISCOVERY:
            //discovery stage
            stage = new DiscoveryPathStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_DISCOVERY);
            break;
        case STAGE_IDX_LINKS:
            //links stage
            stage = new LinksStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_LINKS);
            break;
        case STAGE_IDX_PM:
            //pm stage
            stage = new PMStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_PM);
            break;
        case STAGE_IDX_NVLINK:
            //NVLink stage
            stage = new NVLinkStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_NVLINK);
            break;
        case STAGE_IDX_SPEED_WIDTH:
            //speed / width stage
            stage = new Speed_WidthStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_SPEED_WIDTH);
            break;
        case STAGE_IDX_MLNX_CNTRS:
            //Diagnostic Counters stage
            stage = new MlnxCntrsStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_MLNX_CNTRS);
            break;
        case STAGE_IDX_VIRTUALIZATION:
            //Virtualiztion stage
            stage = new VirtualizationStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_VIRTUALIZATION);
            break;
        case STAGE_IDX_TEMPSENSING:
            //Temperature stage
            stage = new TemperatureSensing(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_TEMPSENSING);
            break;
        case STAGE_IDX_ROUTERS:
            stage = new RoutersStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_ROUTERS);
            break;
        case STAGE_IDX_QOS:
           //QoS stage
           stage = new QoSStage(&this->ibdiag_obj);
           if (!stage)
               ERR_PRINT("Failed to allocate %s stage\n",
                       STAGE_NAME_QOS);
           break;
        case STAGE_IDX_CC:
            // CC stage
            stage = new CongestionControlStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                        STAGE_NAME_CC);
            break;
        case STAGE_IDX_FAST_RECOVERY:
            // fast recovery stage
            stage = new FastRecoveryStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_FAST_RECOVERY);
            break;

        default:
            break;
    }

    IBDIAGNET_RETURN(stage);
}

IBDiagpath::~IBDiagpath()
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN_VOID;
}

//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int IBDiagpath::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;
    int res = IBDiagapp::HandleOption(name, value);

    /* if res is 1 so it's unknown stage to ibdiagapp, need to
     * check for ibdiagpath stages */
    if (res == 1) {
        if (name == OPTION_SRC_LID) {
            lid_t srclid = (lid_t)strtoul(value.c_str(), NULL, 0);
            if (!is_non_neg_num(value) || IsValidLid(srclid)) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                        OPTION_SRC_LID, value.c_str());
                IBDIAGNET_RETURN(1);
            }
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetSrcLid(srclid);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_DEST_LID) {
            lid_t destlid = (lid_t)strtoul(value.c_str(), NULL, 0);
            if (!is_non_neg_num(value) || IsValidLid(destlid)) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                        OPTION_DEST_LID, value.c_str());
                IBDIAGNET_RETURN(1);
            }
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetDestLid(destlid);
                IBDIAGNET_RETURN(0);
          } else if (name == OPTION_DR_PATH) {
            //0,1,8,5,3,1
            regExp drPath("^\\s*(0(,([0-9])+)*)");
            rexMatch *p_rexRes = drPath.apply(value.c_str());
            if (!p_rexRes) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                        OPTION_DR_PATH, value.c_str());
                IBDIAGNET_RETURN(1);
            }
            delete p_rexRes;
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetDRPath(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_ADAPTIVE_ROUTING) {
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetAdaptiveRouting(true);
                IBDIAGNET_RETURN(0);
        } else if (name == OPTION_LIDS_LIST) {
            DiscoveryPathStage *p_pathStage = ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY]);
            if (!p_pathStage->GetAdaptiveRouting()) {
                ERR_PRINT_ARGS("The option \'--%s\' is allowed only if the parameter \'--%s\' is applied before\n",
                        OPTION_LIDS_LIST, OPTION_ADAPTIVE_ROUTING);
                IBDIAGNET_RETURN(1);
            }

            if (p_pathStage->SetLids(value)) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                        OPTION_LIDS_LIST, value.c_str());
                IBDIAGNET_RETURN(1);
            }

            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SHORT_OUTPUT) {
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetShortOutput(true);
                IBDIAGNET_RETURN(0);
        }
    }

    IBDIAGNET_RETURN(res);
}

int IBDiagpath::Init()
{
    IBDIAGNET_ENTER;
    IBDiagapp::Init();

    this->AddDescription(IBDIAGPATH_TOOL_DESC);
    this->AddOptions(OPTION_SRC_LID, OPTION_SRC_LID_SHORT,
                     OPTION_SRC_LID_ARG_DESC, OPTION_SRC_LID_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_DEST_LID, OPTION_DEST_LID_SHORT,
                     OPTION_DEST_LID_ARG_DESC, OPTION_DEST_LID_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_DR_PATH, OPTION_DR_PATH_SHORT,
                     OPTION_DR_PATH_ARG_DESC, OPTION_DR_PATH_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_LIDS_LIST, OPTION_LIDS_LIST_SHORT,
                     OPTION_LIDS_LIST_ARG_DESC, OPTION_LIDS_LIST_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_ADAPTIVE_ROUTING, OPTION_ADAPTIVE_ROUTING_SHORT,
                     OPTION_ADAPTIVE_ROUTING_ARG_DESC, OPTION_ADAPTIVE_ROUTING_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SHORT_OUTPUT, OPTION_SHORT_OUTPUT_SHORT,
                     OPTION_SHORT_OUTPUT_ARG_DESC, OPTION_SHORT_OUTPUT_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN|IBIS_CMD_BOOLEAN_FLAG);

    ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
        TurnOFFDUPGUIDsDetection();
    IBDIAGNET_RETURN(0);
}


#define N_ELEMENTS(array) sizeof(array)/sizeof(array[0])

#define COMPILE_TIME_ASSERT(pred)            \
    switch(0){case 0:case pred:;}

/****************************************************/
int main (int argc, const char **argv)
{
    OutputControl::app_settings.init("ibdiagpath");

    IBDiagpath *app = new IBDiagpath();
    //Any created files will have mode
    //S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH (0666)
     umask(0);

    int rc = app->Run(argc, argv);
    delete app;

    IBDIAGNET_RETURN(rc);
}

